/**
 * Copyright 2013 JogAmp Community. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification, are
 * permitted provided that the following conditions are met:
 *
 *    1. Redistributions of source code must retain the above copyright notice, this list of
 *       conditions and the following disclaimer.
 *
 *    2. Redistributions in binary form must reproduce the above copyright notice, this list
 *       of conditions and the following disclaimer in the documentation and/or other materials
 *       provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY JogAmp Community ``AS IS'' AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL JogAmp Community OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 * ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * The views and conclusions contained in the software and documentation are those of the
 * authors and should not be interpreted as representing official policies, either expressed
 * or implied, of JogAmp Community.
 */
import java.nio.FloatBuffer;

import javax.media.opengl.GL;
import javax.media.opengl.GL2ES2;
import javax.media.opengl.GLAutoDrawable;
import javax.media.opengl.GLCapabilities;
import javax.media.opengl.GLEventListener;
import javax.media.opengl.GLProfile;
import javax.media.opengl.GLUniformData;

import com.jogamp.newt.event.KeyEvent;
import com.jogamp.newt.event.KeyListener;
import com.jogamp.newt.event.MouseEvent;
import com.jogamp.newt.event.MouseListener;
import com.jogamp.newt.event.WindowAdapter;
import com.jogamp.newt.event.WindowEvent;
import com.jogamp.newt.opengl.GLWindow;
import com.jogamp.opengl.util.FPSAnimator;
import com.jogamp.opengl.util.GLArrayDataServer;
import com.jogamp.opengl.util.PMVMatrix;
import com.jogamp.opengl.util.glsl.ShaderCode;
import com.jogamp.opengl.util.glsl.ShaderProgram;
import com.jogamp.opengl.util.glsl.ShaderState;

/***
 * @author Petr Skramovsky
 */
public abstract class Demo implements GLEventListener {
    private ShaderState st;
    private PMVMatrix pmv;
    private GLUniformData pmvUniform;
    private float[] camMatrix;
    private GLArrayDataServer vbo;
    public Camera camera;

    public Demo() {
        camera = new Camera();
        camera.setPosition(0, 0, -10);
        camMatrix = new float[16];
        st = new ShaderState();
        pmv = new PMVMatrix();
        pmv.glMatrixMode(PMVMatrix.GL_PROJECTION);
        pmv.glLoadIdentity();
        pmv.glMatrixMode(PMVMatrix.GL_MODELVIEW);
        pmv.glLoadIdentity();
    }

    private ShaderProgram loadShader(GL2ES2 gl, String name) {
        ShaderCode vp = ShaderCode.create(gl, GL2ES2.GL_VERTEX_SHADER,
                this.getClass(), "shader", "shader/bin", name, true);
        ShaderCode fp = ShaderCode.create(gl, GL2ES2.GL_FRAGMENT_SHADER,
                this.getClass(), "shader", "shader/bin", name, true);
        ShaderProgram sp = new ShaderProgram();
        sp.add(gl, vp, System.err);
        sp.add(gl, fp, System.err);
        sp.link(gl, System.err);
        return sp;
    }

    @Override
    public void init(GLAutoDrawable drawable) {
        GL2ES2 gl = drawable.getGL().getGL2ES2();
        st.setVerbose(true);
        st.attachShaderProgram(gl, loadShader(gl, "simple"), true);
        pmvUniform = new GLUniformData("PMvMvitMatrix", 4, 4,
                pmv.glGetPMvMvitMatrixf()); // pmv.h
        st.ownUniform(pmvUniform);
        vbo = GLArrayDataServer.createGLSLInterleaved(3 + 4, GL.GL_FLOAT,
                false, 3, GL.GL_STATIC_DRAW);
        vbo.addGLSLSubArray("vbo_Vertex", 3, GL.GL_ARRAY_BUFFER);
        vbo.addGLSLSubArray("vbo_Color", 4, GL.GL_ARRAY_BUFFER);
        FloatBuffer vb = (FloatBuffer) vbo.getBuffer();
        vb.put(new float[] { -1.0f, -1.0f, 0.0f });
        vb.put(new float[] { 1.0f, 0.0f, 0.0f, 1.0f });
        vb.put(new float[] { 1.0f, -1.0f, 0.0f });
        vb.put(new float[] { 0.0f, 1.0f, 0.0f, 1.0f });
        vb.put(new float[] { 0.0f, 1.0f, 0.0f });
        vb.put(new float[] { 0.0f, 0.0f, 1.0f, 1.0f });
        vbo.seal(true);
        st.ownAttribute(vbo, true);
        gl.glClearColor(1, 1, 1, 1);
        gl.glEnable(GL2ES2.GL_DEPTH_TEST);
    }

    @Override
    public void dispose(GLAutoDrawable drawable) {
        GL2ES2 gl = drawable.getGL().getGL2ES2();
        st.destroy(gl);
        pmv.destroy();
    }

    @Override
    public void display(GLAutoDrawable drawable) {
        processEvents();
        GL2ES2 gl = drawable.getGL().getGL2ES2();
        gl.glClear(GL2ES2.GL_COLOR_BUFFER_BIT | GL2ES2.GL_DEPTH_BUFFER_BIT);
        pmv.glMatrixMode(PMVMatrix.GL_MODELVIEW);
        pmv.glLoadIdentity();
        pmv.glMultMatrixf(camera.getRotation().toMatrix(camMatrix), 0);
        float[] pos = camera.getPosition();
        pmv.glTranslatef(pos[0], pos[1], pos[2]);
        st.useProgram(gl, true);
        st.uniform(gl, pmvUniform);
        vbo.enableBuffer(gl, true);
        gl.glDrawArrays(GL.GL_TRIANGLES, 0, 3);
        vbo.enableBuffer(gl, false);
        st.useProgram(gl, false);
    }

    @Override
    public void reshape(GLAutoDrawable drawable, int x, int y, int width,
            int height) {
        pmv.glMatrixMode(PMVMatrix.GL_PROJECTION);
        pmv.glLoadIdentity();
        pmv.gluPerspective(45.0f, ((float) width / (float) height) / 1.0f,
                1.0f, 1000.0f);
    }

    public abstract void processEvents();

    public static void main(String[] args) {
        GLWindow w = GLWindow.create(new GLCapabilities(GLProfile
                .get(GLProfile.GL2ES2)));
        final FPSAnimator anim = new FPSAnimator(w, 30, false);
        final EventsListener l = new EventsListener();
        w.addGLEventListener(new Demo() {

            @Override
            public void processEvents() {
                if (l.rx != 0) {
                    camera.rotateY(l.rx * 0.0022f);
                    l.rx = 0;
                }
                if (l.ry != 0) {
                    camera.rotateX(l.ry * 0.0022f);
                    l.ry = 0;
                }

                if (l.w)
                    camera.moveZ(0.25f);
                else if (l.s)
                    camera.moveZ(-0.25f);
                if (l.a)
                    camera.moveX(0.25f);
                else if (l.d)
                    camera.moveX(-0.25f);

            }
        });
        w.addKeyListener(l);
        w.addMouseListener(l);
        w.setSize(640, 640);
        w.setTitle("Quaternion camera demo");
        w.addWindowListener(new WindowAdapter() {
            @Override
            public void windowDestroyNotify(WindowEvent arg0) {
                anim.stop();
            };
        });
        w.setVisible(true);
        anim.start();
    }

    private static class EventsListener implements KeyListener, MouseListener {

        public boolean w, a, s, d;
        public int rx, ry;
        private int px, py;

        public EventsListener() {
            w = a = s = d = false;
        }

        @Override
        public void keyPressed(KeyEvent e) {
            switch (e.getKeyCode()) {
            case KeyEvent.VK_W:
                w = true;
                break;
            case KeyEvent.VK_S:
                s = true;
                break;
            case KeyEvent.VK_A:
                a = true;
                break;
            case KeyEvent.VK_D:
                d = true;
                break;
            case KeyEvent.VK_ESCAPE:
                System.exit(0);
                break;
            }
        }

        @Override
        public void keyReleased(KeyEvent e) {

        }

        @Override
        public void keyTyped(KeyEvent e) {
            switch (e.getKeyCode()) {
            case KeyEvent.VK_W:
                w = false;
                break;
            case KeyEvent.VK_S:
                s = false;
                break;
            case KeyEvent.VK_A:
                a = false;
                break;
            case KeyEvent.VK_D:
                d = false;
                break;
            }
        }

        @Override
        public void mouseClicked(MouseEvent e) {
        }

        @Override
        public void mouseEntered(MouseEvent e) {
        }

        @Override
        public void mouseExited(MouseEvent e) {
            rx = ry = px = py = 0;
        }

        @Override
        public void mousePressed(MouseEvent e) {
        }

        @Override
        public void mouseReleased(MouseEvent e) {
            rx = ry = px = py = 0;
        }

        @Override
        public void mouseMoved(MouseEvent e) {
        }

        @Override
        public void mouseDragged(MouseEvent e) {
            if (px != 0 && py != 0) {
                rx += px - e.getX();
                ry += py - e.getY();
            }
            px = e.getX();
            py = e.getY();
        }

        @Override
        public void mouseWheelMoved(MouseEvent e) {
        }

    }
}
